﻿Class MainWindow 

    Enum State
        Go
        Back
    End Enum

    Dim BasePixel As Double
    Dim Distance As Double
    Dim SettingComplete As Boolean
    Dim CurrentSize As Decimal
    Dim TestState As State
    Dim CorrectCount As Integer
    Dim DispRing As Integer

    Private Sub MainWindow_Loaded(sender As Object, e As RoutedEventArgs) Handles Me.Loaded

        BasePixel = 200
        Distance = 1
        SettingComplete = False

    End Sub

    Sub ExitProgram(sender As Object, e As ExecutedRoutedEventArgs)

        Me.Close()

    End Sub

    Sub TestSetting(sender As Object, e As ExecutedRoutedEventArgs)

        Dim SettingDialog As SettingWindow = New SettingWindow()
        SettingDialog.BasePixel = BasePixel
        SettingDialog.Distance = Distance
        If SettingDialog.ShowDialog() Then
            BasePixel = SettingDialog.BasePixel
            Distance = SettingDialog.Distance
            SettingComplete = True
        End If

    End Sub

    Sub StartCheck(sender As Object, e As ExecutedRoutedEventArgs)

        If Not SettingComplete Then
            MessageBox.Show("先に検査設定を行ってください", Me.Title, MessageBoxButton.OK, MessageBoxImage.Error)
        Else
            CurrentSize = 0.1
            TestState = State.Go
            CorrectCount = 0
            DispRing = 0
            DispRandort(CurrentSize, Distance)
            ButtonFalse.IsEnabled = True
            ButtonTrue.IsEnabled = True
        End If

    End Sub

    Sub DispRandort(s As Double, d As Double)

        If CurrentSize < 0.1 Then
            Exit Sub
        End If

        Label1.Content = String.Format("視力: {0:0.0}", s)

        Dim dr As Double = 3 * d * BasePixel / (100 * s)
        Dim x1 As Double = dr
        Dim x2 As Double = x1 - (dr / 5)
        Dim y1 As Double = (2 / 5) * dr
        Dim y2 As Double = y1 + (1 / 5) * dr

        Dim p1 As Point = New Point(x1, y1)
        Dim p2 As Point = New Point(x1, y2)
        Dim p3 As Point = New Point(x2, y2)
        Dim p4 As Point = New Point(x2, y1)

        Dim Figure As PathFigure = New PathFigure() With {.StartPoint = p1}
        Figure.Segments.Add(New ArcSegment(p2, New Size(dr / 2, dr / 2), 0, True, SweepDirection.Counterclockwise, True))
        Figure.Segments.Add(New LineSegment(p3, True))
        Figure.Segments.Add(New ArcSegment(p4, New Size((dr * 0.6) / 2, (dr * 0.6) / 2), 0, True, SweepDirection.Clockwise, True))
        Dim MyGeo As PathGeometry = New PathGeometry()
        MyGeo.Figures.Add(Figure)

        Dim MyRandom As Random = New Random()
        Dim MyPath As Path = New Path() With {.Data = MyGeo}
        MyPath.Width = dr
        MyPath.Height = dr
        MyPath.RenderTransformOrigin = New Point(0.5, 0.5)
        MyPath.RenderTransform = New RotateTransform(MyRandom.Next(8) * 45, MyPath.ActualWidth / 2, MyPath.ActualHeight / 2)
        MyPath.Fill = Brushes.Black

        Canvas.SetLeft(MyPath, (Canvas1.ActualWidth - dr) / 2)
        Canvas.SetTop(MyPath, (Canvas1.ActualHeight - dr) / 2)

        Canvas1.Children.Clear()
        Canvas1.Children.Add(MyPath)

    End Sub

    Private Sub Button_Click(sender As Object, e As RoutedEventArgs)

        Select Case TestState
            Case State.Go
                If e.OriginalSource Is ButtonTrue Then
                    CurrentSize += 0.1  ' 次の視力に進む
                    If CurrentSize = 1 Then
                        TestState = State.Back
                        DispRing = 0
                        CorrectCount = 0
                    End If
                Else
                    TestState = State.Back
                    DispRing = 1    ' 1つリングを表示して正解は0という状態である
                    CorrectCount = 0
                End If
            Case State.Back
                CorrectCount += IIf(e.OriginalSource Is ButtonTrue, 1, 0)
                DispRing += 1
                If DispRing = 3 Then
                    If CorrectCount > 1 Then    ' 3つリングを表示して2つ正解であればその視力とする
                        MessageBox.Show("視力は" & CurrentSize & "です", Me.Title, MessageBoxButton.OK, MessageBoxImage.Information)
                    Else
                        ' 一つ前のリングに戻す
                        CurrentSize -= 0.1
                        If CurrentSize < 0.1 Then
                            MessageBox.Show("あなたの視力は0.1未満です", Me.Title, MessageBoxButton.OK, MessageBoxImage.Information)
                        End If
                        DispRing = 0
                        CorrectCount = 0
                    End If
                End If
        End Select

        DispRandort(CurrentSize, Distance)

    End Sub
End Class
