﻿Imports System.Windows.Media.Effects
Imports Microsoft.Win32
Class MainWindow

    Dim HBeziers(,) As Point    ' 水平方向のベジエ曲線
    Dim VBeziers(,) As Point    ' 垂直方向のベジエ曲線

    Dim Dpoint As Point ' ドラッグの際のオフセット
    Dim IsDrag As Boolean   ' 現在ドラッグ中かどうか

    Dim ImageWidth As Integer   ' 元イメージの幅
    Dim ImageHeight As Integer  ' 元イメージの高さ

    Dim IsAlreadyCompleted As Boolean ' 完成済みとなったか?
    Dim PuzzleCells As List(Of PuzzleCell)   ' セルのコレクション

    Private Sub MainWindow_Loaded(sender As Object, e As RoutedEventArgs) Handles MyBase.Loaded
        InitBezier()
    End Sub

    Sub InitBezier()

        ' 左から右への水平のベジエ曲線のポイント
        Dim Beziers(,) As Point = New Point(3, 9) {{New Point(0, 0), New Point(24, -12), New Point(108, -24), New Point(74, 24), New Point(49, 78), New Point(168, 78), New Point(143, 24), New Point(108, -24), New Point(191, -12), New Point(216, 0)},
                           {New Point(0, 0), New Point(24, -12), New Point(119, -23), New Point(79, 24), New Point(64, 78), New Point(172, 80), New Point(150, 24), New Point(120, -23), New Point(191, -12), New Point(216, 0)},
                           {New Point(0, 0), New Point(35, 0), New Point(132, -26), New Point(92, 23), New Point(79, 77), New Point(168, 78), New Point(143, 18), New Point(129, -25), New Point(191, 3), New Point(216, 0)},
                           {New Point(0, 0), New Point(24, -12), New Point(112, -4), New Point(81, 24), New Point(54, 86), New Point(158, 82), New Point(133, 25), New Point(108, -14), New Point(190, -3), New Point(216, 0)}}

        ' 実際に利用するベジエ曲線のポイント()
        HBeziers = New Point(16, 9) {}
        VBeziers = New Point(16, 9) {}

        ' 初期パターンをHBeziers(0)～HBeziers(3)に、左右反転パターンをHBeziers(4)～HBeziers(7)に作成
        For i As Integer = 0 To 3
            For p As Integer = 0 To 9
                HBeziers(i, p) = Beziers(i, p)
                HBeziers(i + 4, p) = New Point(216 - Beziers(i, 9 - p).X, Beziers(i, 9 - p).Y)
            Next
        Next

        ' 上下反転パターンをHBeziers(8)～HBeziers(15)に作成
        For i As Integer = 0 To 7
            For p As Integer = 0 To 9
                HBeziers(i + 8, p) = New Point(HBeziers(i, p).X, -HBeziers(i, p).Y)
            Next
        Next

        ' 直線パターンを作成(最後の点だけ216,0)
        For p As Integer = 0 To 9
            HBeziers(16, p) = New Point(If(p = 9, 216, 0), 0)
        Next

        ' HBeziers(0)～HBeziers(16)からVBeziers(0)～HBeziers(16)を作成
        For i As Integer = 0 To 16
            For p As Integer = 0 To 9
                VBeziers(i, p) = New Point(HBeziers(i, p).Y, HBeziers(i, p).X)
            Next
        Next

    End Sub

    Sub MakePuzzleBoard()

        Dim MyRandom As Random = New Random()

        Dim HPattern As Integer(,) = New Integer(PuzzleCell.DivX - 1, PuzzleCell.DivY) {}
        Dim VPattern As Integer(,) = New Integer(PuzzleCell.DivX, PuzzleCell.DivY - 1) {}

        ' パズル面の区切り線の種類を決める
        For y As Integer = 0 To PuzzleCell.DivY
            For x As Integer = 0 To PuzzleCell.DivX
                If x <> PuzzleCell.DivX Then
                    ' 水平線は上端と下端は直線
                    HPattern(x, y) = If(y = 0 Or y = PuzzleCell.DivY, 16, MyRandom.Next(16))
                End If
                If y <> PuzzleCell.DivY Then
                    ' 垂直線は左端と右端は直線
                    VPattern(x, y) = If(x = 0 Or x = PuzzleCell.DivX, 16, MyRandom.Next(16))
                End If
            Next
        Next

        ' PuzzleCellを作成しキャンバスにPathを追加

        Canvas1.Children.Clear()
        PuzzleCells = New List(Of PuzzleCell)
        For i As Integer = 0 To PuzzleCell.DivX * PuzzleCell.DivY - 1
            Dim c As PuzzleCell = New PuzzleCell(i, i)
            PuzzleCells.Add(c)
            SetPiecePath(c, HPattern(c.GetX, c.GetY), HPattern(c.GetX, c.GetY + 1), VPattern(c.GetX, c.GetY), VPattern(c.GetX + 1, c.GetY))
            Canvas1.Children.Add(c.PiecePath)
        Next

    End Sub

    ' PuzzleCellに必要なPathを作成してセット
    Private Function SetPiecePath(c As PuzzleCell, HPattern1 As Integer, HPattern2 As Integer, VPattern1 As Integer, VPattern2 As Integer) As Integer

        Dim RatioPoint = Function(pp As Point)
                             Return New Point(pp.X / 216 * PuzzleCell.CellSize, pp.Y / 216 * PuzzleCell.CellSize)
                         End Function

        Dim Figure As PathFigure = New PathFigure()

        ' 始点を設定
        Figure.StartPoint = RatioPoint(HBeziers(HPattern1, 0))
        Dim BSegment As PolyBezierSegment = New PolyBezierSegment()

        ' 左から右への上の線
        For i = 1 To 9
            BSegment.Points.Add(RatioPoint(HBeziers(HPattern1, i)))
        Next

        ' 右上から右下への線
        For i = 1 To 9
            BSegment.Points.Add(RatioPoint(New Point(VBeziers(VPattern2, i).X + 216, VBeziers(VPattern2, i).Y)))
        Next

        ' 右下から左下への線
        For i = 8 To 0 Step -1
            BSegment.Points.Add(RatioPoint(New Point(HBeziers(HPattern2, i).X, HBeziers(HPattern2, i).Y + 216)))
        Next

        ' 左下から左上への線
        For i = 8 To 0 Step -1
            BSegment.Points.Add(RatioPoint(New Point(VBeziers(VPattern1, i).X, VBeziers(VPattern1, i).Y)))
        Next

        Figure.Segments.Add(BSegment)

        Dim Geo As PathGeometry = New PathGeometry()
        Geo.Figures.Add(Figure)

        ' イメージブラシを作成

        Dim ViewBoxRect As Rect = Geo.GetRenderBounds(New Pen(Brushes.Transparent, 0.5))

        Dim MyBrush As ImageBrush = New ImageBrush()
        MyBrush.ViewboxUnits = BrushMappingMode.Absolute
        MyBrush.ImageSource = Image1.Source
        MyBrush.Viewbox = New Rect(c.GetLeftOffset() + ViewBoxRect.X, c.GetTopOffset() + ViewBoxRect.Y, PuzzleCell.CellSize, PuzzleCell.CellSize)
        MyBrush.Stretch = Stretch.None
        MyBrush.AlignmentX = AlignmentX.Left
        MyBrush.AlignmentY = AlignmentY.Top

        Dim Piece As Path = New Path()
        Piece.Stroke = Brushes.Black
        Piece.StrokeThickness = 0.5
        Piece.Data = Geo
        Piece.Fill = MyBrush
        Piece.Tag = c   ' PathからCellをたどれるように

        c.PiecePath = Piece

        Return 0

    End Function

    Private Sub OpenCommandBinding_Executed(sender As Object, e As ExecutedRoutedEventArgs)

        Dim MyDialog As OpenFileDialog = New OpenFileDialog()
        MyDialog.Filter = "イメージファイル(*.bmp *.jpg *.jpeg *.png)|*.bmp;*.jpg;*.jpeg;*.png|すべてのファイル(*.*)|*.*"
        MyDialog.CheckFileExists = True

        If MyDialog.ShowDialog() Then

            Dim PuzzleBitmap As BitmapImage = New BitmapImage()
            Dim s As System.IO.FileStream = New System.IO.FileStream(MyDialog.FileName, System.IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read)
            Dim d As BitmapDecoder = BitmapDecoder.Create(s, BitmapCreateOptions.None, BitmapCacheOption.Default)

            Image1.Source = d.Frames(0)
            ImageWidth = Image1.Source.Width
            ImageHeight = Image1.Source.Height
            SizeSelect()
            s.Close()

        End If

    End Sub

    Private Sub NewCommandBinding_CanExecute(sender As Object, e As CanExecuteRoutedEventArgs)

        e.CanExecute = Not (Image1.Source Is Nothing)

    End Sub

    Private Sub CloseCommandBinding_Executed(sender As Object, e As ExecutedRoutedEventArgs)
        Me.Close()
    End Sub

    Private Sub NewCommandBinding_Executed(sender As Object, e As ExecutedRoutedEventArgs)
        SizeSelect()
    End Sub

    ' サイズ設定ウィンドウを開いてサイズを決定しセルを配置
    Sub SizeSelect()

        Dim MySettingWindow As SettingWindow = New SettingWindow()
        MySettingWindow.WidthP = Image1.Source.Width
        MySettingWindow.HeightP = Image1.Source.Height

        If MySettingWindow.ShowDialog() Then

            Dim myrandom As Random = New Random()

            Dim w As Integer = MySettingWindow.PieceSize
            PuzzleCell.CellSize = w
            PuzzleCell.DivX = ImageWidth \ w
            PuzzleCell.DivY = ImageHeight \ w

            MakePuzzleBoard()

            ' キャンバス上にセルをランダムに配置する
            For Each c As PuzzleCell In PuzzleCells
                c.SetCanvasPosition(New Point(myrandom.Next(Canvas1.ActualWidth - w),
                                              myrandom.Next(Canvas1.ActualHeight - w)))
            Next
            IsAlreadyCompleted = False
        End If

    End Sub

    ' パズルが完成したか(すべてのセルが同じグループに属しているか?)
    Function IsComplete() As Boolean

        Dim g As Integer = PuzzleCells(0).GroupId

        For Each cell As PuzzleCell In PuzzleCells
            If cell.GroupId <> g Then
                Return False
            End If
        Next

        Return True

    End Function

    ' 指定されたパスを一番上に移動する
    Sub MoveToTop(p As Path)

        Dim x As Integer = 0
        For Each m As FrameworkElement In Canvas1.Children
            Dim s As Integer = Canvas.GetZIndex(m)
            x = Math.Max(s, x)
        Next
        Canvas.SetZIndex(p, x + 1)

    End Sub

    ' c1のセルからみた正しいc2のセルの位置を求める
    Function AccuratePosition(c1 As PuzzleCell, c2 As PuzzleCell) As Point

        Dim p As Point = c1.CanvasPosition()
        Return New Point(p.X + c2.GetLeftOffset - c1.GetLeftOffset, p.Y + c2.GetTopOffset - c1.GetTopOffset)

    End Function

    ' cのセルの上下左右のセルが近接している場合は同じグループにして位置も移動する
    Sub CheckMatch(c As PuzzleCell)

        Dim IsNearPath = Function(c1 As PuzzleCell, c2 As PuzzleCell) As Boolean
                             Dim dx As Double = AccuratePosition(c1, c2).X - c2.CanvasPosition.X
                             Dim dy As Double = AccuratePosition(c1, c2).Y - c2.CanvasPosition.Y
                             Dim d As Double = Math.Sqrt(dx ^ 2 + dy ^ 2)
                             If d < 6 Then  ' セルが正しい位置より6ピクセル以内の場合は近接と判断
                                 Return True
                             Else
                                 Return False
                             End If
                         End Function

        Dim CheckJoin = Sub(c1 As PuzzleCell, c2 As PuzzleCell)
                            If c1.GroupId <> c2.GroupId AndAlso IsNearPath(c1, c2) Then
                                JoinGroup(c1, c2)
                            End If
                        End Sub

        If Not c.IsLeftBorder Then  ' 左のセルをチェック
            CheckJoin(c, SearchCell(c.GetX - 1, c.GetY))
        End If
        If Not c.IsRightBorder Then ' 右のセルをチェック
            CheckJoin(c, SearchCell(c.GetX + 1, c.GetY))
        End If
        If Not c.IsTopBorder Then   ' 上のセルをチェック
            CheckJoin(c, SearchCell(c.GetX, c.GetY - 1))
        End If
        If Not c.IsBottomBorder Then    ' 下のセルをチェック
            CheckJoin(c, SearchCell(c.GetX, c.GetY + 1))
        End If
    End Sub

    ' xとyの位置を持つセルを返す
    Function SearchCell(x As Integer, y As Integer) As PuzzleCell

        For Each c As PuzzleCell In PuzzleCells
            If c.GetX = x And c.GetY = y Then
                Return c
            End If
        Next
        Return Nothing

    End Function

    ' c1をc2と同じグループにして、c1の位置を調整する
    Sub JoinGroup(c1 As PuzzleCell, c2 As PuzzleCell)

        Dim z As Integer = Math.Max(Canvas.GetZIndex(c1.PiecePath), Canvas.GetZIndex(c2.PiecePath))
        ' グループIDを取得
        Dim g1 As Integer = c1.GroupId
        Dim g2 As Integer = c2.GroupId

        If g1 = g2 Then
            ' すでに同じ場合はなにもしない
            Return
        End If

        ' c1と同じグループのセルの位置を調整する。c2と同じグループのセルはZindexを調整する
        For Each c As PuzzleCell In PuzzleCells
            If c.GroupId = g1 Then
                Dim p As Point = AccuratePosition(c2, c)
                c.SetCanvasPosition(p)
                Canvas.SetZIndex(c.PiecePath, z)
            ElseIf c.GroupId = g2 Then
                c.GroupId = g1
                Canvas.SetZIndex(c.PiecePath, z)
            End If
        Next

    End Sub

    Private Sub Canvas1_MouseLeftButtonDown(sender As Object, e As MouseButtonEventArgs)

        If TypeOf e.OriginalSource Is Path Then
            Dim target As Path = CType(e.OriginalSource, Path)

            MoveToTop(target)
            ' 同じグループのセルは同じZindexを設定
            Dim z As Integer = Canvas.GetZIndex(target)
            Dim g As Integer = CType(target.Tag, PuzzleCell).GroupId

            For Each c As PuzzleCell In PuzzleCells
                If c.GroupId = g Then
                    Canvas.SetZIndex(c.PiecePath, z)
                End If
            Next
            Dpoint = e.GetPosition(target)
            IsDrag = True
            Mouse.Capture(target)
        End If

    End Sub

    Private Sub Canvas1_MouseMove(sender As Object, e As MouseEventArgs)

        If IsDrag And TypeOf e.OriginalSource Is Path Then
            Dim target As Path = CType(e.OriginalSource, Path)
            Dim p As Point = e.GetPosition(Canvas1)
            Canvas.SetLeft(target, p.X - Dpoint.X)
            Canvas.SetTop(target, p.Y - Dpoint.Y)
            Dim g As Integer = CType(target.Tag, PuzzleCell).GroupId

            ' 同じグループのセルは同時に移動
            For Each c As PuzzleCell In PuzzleCells
                If c.GroupId = g Then
                    c.SetCanvasPosition(AccuratePosition(target.Tag, c))
                End If
            Next
        End If

    End Sub

    Private Sub Canvas1_MouseLeftButtonUp(sender As Object, e As MouseButtonEventArgs)

        If IsDrag And TypeOf e.OriginalSource Is Path Then
            IsDrag = False
            Mouse.Capture(Nothing)
            Dim target As Path = CType(e.OriginalSource, Path)
            Dim g As Integer = CType(target.Tag, PuzzleCell).GroupId

            ' ドラッグしたセルと同じグループのセルに隣接するセルをチェック
            For Each c As PuzzleCell In PuzzleCells
                If c.GroupId = g Then
                    CheckMatch(c)
                End If
            Next
        End If
        If IsComplete() And Not IsAlreadyCompleted Then
            MessageBox.Show("完成おめでとう", Me.Title, MessageBoxButton.OK, MessageBoxImage.Information)
            IsAlreadyCompleted = True
        End If

    End Sub
End Class

Public Class PuzzleCell
    Public Index As Integer    ' セルのインデックス
    Public Shared DivX As Integer ' 1行あたりのセル数
    Public Shared DivY As Integer ' 行数
    Public Shared CellSize As Double ' セルのサイズ(pixel数)
    Public PiecePath As Path ' 対応しているパス
    Public GroupId As Integer ' 所属するグループID

    Sub New(i As Integer, g As Integer)
        Index = i
        GroupId = g
    End Sub

    Function GetX() As Integer
        Return Index Mod DivX
    End Function

    Function GetY() As Integer
        Return Index \ DivX
    End Function

    Function IsLeftBorder() As Boolean
        Return GetX() = 0
    End Function

    Function IsRightBorder() As Boolean
        Return GetX() = DivX - 1
    End Function

    Function IsTopBorder() As Boolean
        Return Index < DivX
    End Function

    Function IsBottomBorder() As Boolean
        Return GetY() = DivY - 1
    End Function

    Function CanvasPosition() As Point
        If PiecePath IsNot Nothing Then
            Return New Point(Canvas.GetLeft(PiecePath), Canvas.GetTop(PiecePath))
        Else
            Return Nothing
        End If
    End Function

    Sub SetCanvasPosition(p As Point)
        If PiecePath IsNot Nothing Then
            Canvas.SetLeft(PiecePath, p.X)
            Canvas.SetTop(PiecePath, p.Y)
        End If
    End Sub

    Function GetLeftOffset() As Double
        Return CellSize * GetX()
    End Function

    Function GetTopOffset() As Double
        Return CellSize * GetY()
    End Function

End Class
