﻿Imports System.Windows.Media.Animation
Class MainWindow

    Dim PieceNumber() As Integer = New Integer(15) {}
    Dim Hole As Integer
    Dim Mode As Integer
    Dim IsMoving As Boolean

    ' Placeの位置にPieceNumber()が示すパネルを表示する
    Sub DispPanel(Place As Integer)

        Dim PanelBitmap As BitmapImage

        If PieceNumber(Place) = 15 Then
            PanelBitmap = Nothing
        Else
            Dim BitmapUril As Uri = New Uri(String.Format("/Images/Panel{0:00}.bmp", PieceNumber(Place) + 1), UriKind.Relative)
            PanelBitmap = New BitmapImage(BitmapUril)
        End If

        CType(Canvas1.FindName("Image" & Place), Image).Source = PanelBitmap

    End Sub

    Private Sub MainWindow_Loaded(sender As Object, e As RoutedEventArgs) Handles Me.Loaded

        NewGame(Nothing, Nothing)

    End Sub

    Private Sub NewGame(sender As Object, r As RoutedEventArgs)

        ' パネルを初期化
        Hole = 15
        For i As Integer = 0 To 15
            PieceNumber(i) = i
        Next

        ' 問題を作る
        Shuffle(20)

        ' パネルを表示する
        For i As Integer = 0 To 15
            DispPanel(i)
        Next

        Mode = 1

    End Sub

    ' 15パズル完成の判定
    Function IsComplete() As Boolean

        Dim Complete As Boolean
        If Mode = 0 Then
            ' 一度完成したらもう表示しない
            IsComplete = False
            Exit Function
        End If

        Complete = True
        For i = 0 To 14
            If PieceNumber(i) <> i Then
                Complete = False
            End If
        Next
        Return Complete

    End Function

    ' ランダムにパネルを移動して問題を作る
    Sub Shuffle(Count As Integer)

        Dim MyRandom As Random = New Random()
        Dim LastN As Integer = -1

        For i As Integer = 1 To Count
            Dim n As Integer = MyRandom.Next(4)
            Select Case n
                Case 0  '上
                    If Hole > 3 And LastN <> 2 Then
                        ShuffleMove(Hole - 4)
                        LastN = 0
                    Else
                        i = i - 1
                    End If
                Case 1  ' 左
                    If Hole Mod 4 <> 0 And LastN <> 3 Then
                        ShuffleMove(Hole - 1)
                        LastN = 1
                    Else
                        i = i - 1
                    End If
                Case 2  ' 下
                    If Hole < 12 And LastN <> 0 Then
                        ShuffleMove(Hole + 4)
                        LastN = 2
                    Else
                        i = i - 1
                    End If
                Case 3  ' 右
                    If Hole Mod 4 <> 3 And LastN <> 1 Then
                        ShuffleMove(Hole + 1)
                        LastN = 3
                    Else
                        i = i - 1
                    End If
            End Select
        Next

    End Sub

    ' 新しい穴の位置をNumに変更
    Sub ShuffleMove(Num)

        PieceNumber(Hole) = PieceNumber(Num)
        PieceNumber(Num) = 15
        Hole = Num

    End Sub

    Private Sub Canvas1_MouseLeftButtonDown(sender As Object, e As MouseButtonEventArgs)

        ' アニメ中は移動しない
        If IsMoving Or Not TypeOf e.OriginalSource Is Image Then
            Return
        End If

        ' パネルのインデックスを取得する
        Dim TouchImage As Image = CType(e.OriginalSource, Image)
        Dim Index As Integer = TouchImage.Tag

        Dim ToLeft As Double = Canvas.GetLeft(TouchImage)
        Dim ToTop As Double = Canvas.GetTop(TouchImage)

        ' アニメーションを作成するラムダ式
        Dim MakeAnimation = Function(f As Double, t As Double, p As String) As DoubleAnimationUsingKeyFrames
                                Dim MoveAnimation As DoubleAnimationUsingKeyFrames = New DoubleAnimationUsingKeyFrames()
                                MoveAnimation.KeyFrames.Add(New LinearDoubleKeyFrame(f, KeyTime.FromTimeSpan(New TimeSpan(0))))
                                MoveAnimation.KeyFrames.Add(New LinearDoubleKeyFrame(t, KeyTime.FromTimeSpan(New TimeSpan(0, 0, 0, 0, 250))))
                                MoveAnimation.KeyFrames.Add(New LinearDoubleKeyFrame(t, KeyTime.FromTimeSpan(New TimeSpan(0, 0, 0, 0, 300))))
                                MoveAnimation.Duration = New Duration(New TimeSpan(0, 0, 0, 0, 300))
                                MoveAnimation.SetValue(Storyboard.TargetProperty, TouchImage)
                                MoveAnimation.SetValue(Storyboard.TargetPropertyProperty, New PropertyPath(p))
                                MoveAnimation.FillBehavior = FillBehavior.Stop
                                Return MoveAnimation
                            End Function

        ' 移動可能かどうかを調べる
        If (Index = Hole + 1) And (Hole Mod 4 <> 3) Then
            ToLeft -= 64
        ElseIf (Index = Hole - 1) And (Hole Mod 4 <> 0) Then
            ToLeft += 64
        ElseIf Index = Hole + 4 Then
            ToTop -= 64
        ElseIf Index = Hole - 4 Then
            ToTop += 64
        Else
            Exit Sub
        End If

        IsMoving = True

        ' 水平方向アニメーション
        Dim LeftAnimation As DoubleAnimationUsingKeyFrames = MakeAnimation(Canvas.GetLeft(TouchImage), ToLeft, "(Canvas.Left)")

        ' 垂直方向アニメーション
        Dim TopAnimation As DoubleAnimationUsingKeyFrames = MakeAnimation(Canvas.GetTop(TouchImage), ToTop, "(Canvas.Top)")
        AddHandler TopAnimation.CurrentStateInvalidated, AddressOf LeftAnimation_Completed

        Dim ImageStoryboard As Storyboard = New Storyboard()
        ImageStoryboard.Children.Add(LeftAnimation)
        ImageStoryboard.Children.Add(TopAnimation)

        ImageStoryboard.Begin()

    End Sub

    ' アニメーション完了時の処理
    Private Sub LeftAnimation_Completed(sender As Object, e As EventArgs)

        Dim MyClock As Clock = CType(sender, Clock)

        If MyClock.CurrentState = ClockState.Stopped Then
            Dim Index As Integer = CType(Storyboard.GetTarget(MyClock.Timeline), Image).Tag
            Dim x As Integer = Hole ' 移動前の穴の位置を保存

            ShuffleMove(Index)

            DispPanel(Hole)
            DispPanel(x)

            If IsComplete() Then
                MessageBox.Show("完成おめでとう!", Me.Title, MessageBoxButton.OK, MessageBoxImage.Information)
                Mode = 0 ' 一度完成したらもうおめでとうを表示しない
            End If

            IsMoving = False
        End If

    End Sub
End Class
