﻿Imports System.Windows.Media.Animation

Enum OpenMode
    Open1   ' 1枚目が開く
    Open2   ' 2枚目が開く
    Close   ' カードを閉じる
    OverOpen    ' ゲームオーバー後にカードを見せる
End Enum

Class MainWindow

    Dim Open1, Open2 As Integer
    Dim Life As Integer
    Dim Cards() As Integer
    Dim Moving As Boolean

    Private Sub Image_MouseLeftButtonDown(sender As Object, e As MouseButtonEventArgs)

        ' アニメ中は処理しない
        If Moving Then
            Exit Sub
        End If

        ' カードのインデックスを取得
        Dim Index As Integer = CType(e.OriginalSource, Image).Tag

        If Cards(Index) = -1 Then   ' すでに開いたカード
            Exit Sub
        End If

        If Open1 = -1 Then
            ' 1枚目を開いた
            Open1 = Index
            OpenAnime(e.OriginalSource, OpenMode.Open1)
        ElseIf Open1 <> Index Then
            ' 2枚目を開いた
            Open2 = Index
            OpenAnime(e.OriginalSource, OpenMode.Open2)
        End If

    End Sub

    ' カードのアニメーションを実現
    Private Sub OpenAnime(AnimeImage As Image, Mode As OpenMode)

        Dim MyStoryBoard As Storyboard = New Storyboard()
        Dim ScaleAnimation = Function(f As Double, t As Double, d As Integer, b As Integer) As DoubleAnimation
                                 Dim MyAnimation As DoubleAnimation = New DoubleAnimation() With {.From = f, .To = t, .Duration = New Duration(New TimeSpan(0, 0, 0, 0, d)), .BeginTime = New TimeSpan(0, 0, 0, 0, b)}
                                 MyAnimation.FillBehavior = FillBehavior.Stop
                                 Storyboard.SetTarget(MyAnimation, AnimeImage)
                                 Storyboard.SetTargetProperty(MyAnimation, New PropertyPath("RenderTransform.ScaleX"))
                                 Return MyAnimation
                             End Function

        Dim DoubleAnimation1 As DoubleAnimation = ScaleAnimation(1, 0, 150, 0)
        If Mode = OpenMode.Close Then
            AddHandler DoubleAnimation1.CurrentStateInvalidated, AddressOf CloseChange
        Else
            AddHandler DoubleAnimation1.CurrentStateInvalidated, AddressOf Openchange
        End If

        Dim DoubleAnimation2 As DoubleAnimation = ScaleAnimation(0, 1, 150, 150)
        Dim DoubleAnimation3 As DoubleAnimation = ScaleAnimation(1, 1, 150, 300)

        MyStoryBoard.Children.Add(DoubleAnimation1)
        MyStoryBoard.Children.Add(DoubleAnimation2)
        MyStoryBoard.Children.Add(DoubleAnimation3)
        If Mode = OpenMode.Open2 Then
            AddHandler DoubleAnimation3.Completed, AddressOf OpenCompleted2
        ElseIf Mode = OpenMode.OverOpen Then
            AddHandler DoubleAnimation3.Completed, AddressOf OverCompleted
        ElseIf Mode = OpenMode.Open1 Then
            AddHandler DoubleAnimation3.Completed, AddressOf OpenCompleted
        End If

        Moving = True
        MyStoryBoard.Begin()

    End Sub

    Sub InitGame()

        Open1 = -1
        Open2 = -2
        Life = 6
        DispLife()
        Shuffle()

    End Sub

    Sub Shuffle()

        Cards = New Integer(15) {}
        ' カードを初期化
        For i = 0 To 15
            Cards(i) = i \ 2
        Next

        ' カードを混ぜる
        Dim MyRandom As Random = New Random()
        For i = 0 To 15
            Dim k As Integer = Cards(i)
            Dim n As Integer = MyRandom.Next(16)
            Cards(i) = Cards(n)
            Cards(n) = k
        Next

    End Sub

    Sub CloseChange(sender As Object, e As EventArgs)
        ImageChange(sender, False)
    End Sub

    Sub Openchange(sender As Object, e As EventArgs)
        ImageChange(sender, True)
    End Sub

    Sub ImageChange(sender As Object, IsFace As Boolean)

        Dim MyClock = CType(sender, Clock)
        If MyClock.CurrentState = ClockState.Stopped Then
            Dim ImageIndex As Integer = CType(Storyboard.GetTarget(MyClock.Timeline), Image).Tag
            Dim Index As Integer = If(IsFace, Cards(ImageIndex), -1)
            Dim FaceImage As ImageSource = New BitmapImage(New Uri("/Images/Card" & Index + 1 & ".png", UriKind.Relative))
            CType(Storyboard.GetTarget(MyClock.Timeline), Image).Source = FaceImage
        End If

    End Sub

    Function IsCompleted() As Boolean

        Dim Result As Boolean = True
        For Each c As Integer In Cards
            If c <> -1 Then
                Result = False
            End If
        Next
        Return Result

    End Function

    Sub OpenCompleted2(sender As Object, e As EventArgs)

        If Cards(Open1) = Cards(Open2) Then ' 2枚開いたカードが一致
            Cards(Open1) = -1
            Cards(Open2) = -1
            CType(Grid1.FindName("Image" & Open1 + 1), Image).Source = Nothing
            CType(Grid1.FindName("Image" & Open2 + 1), Image).Source = Nothing
            Moving = False
            If IsCompleted() Then
                MessageBox.Show("クリアおめでとう", Me.Title, MessageBoxButton.OK, MessageBoxImage.Information)
            End If
        Else
            Life -= 1
            DispLife()

            If Life = 0 Then
                Moving = True
                Dim dflag As Boolean = False
                For i As Integer = 0 To 15
                    ' カードを表にする
                    If Cards(i) <> -1 Then
                        OpenAnime(Grid1.FindName("Image" & i + 1), If(dflag, OpenMode.Open1, OpenMode.OverOpen))
                        dflag = True
                    End If
                Next
            Else
                OpenAnime(Grid1.FindName("Image" & Open1 + 1), OpenMode.Close)
                OpenAnime(Grid1.FindName("Image" & Open2 + 1), OpenMode.Close)
                Moving = False
            End If
        End If
        Open1 = -1
        Open2 = -2

    End Sub

    Sub OverCompleted(sender As Object, e As EventArgs)

        For i As Integer = 0 To 15
            Cards(i) = -1
        Next
        MessageBox.Show("ゲームオーバーです", Me.Title, MessageBoxButton.OK, MessageBoxImage.Error)

    End Sub

    Sub DispLife()

        For i As Integer = 1 To 6
            Dim HeartImage As Image = StackPanel1.FindName("ImageHeart" & i)
            If HeartImage IsNot Nothing Then
                If i > Life Then
                    HeartImage.Visibility = Visibility.Hidden
                Else
                    HeartImage.Visibility = Visibility.Visible
                End If
            End If
        Next

    End Sub

    Sub OpenCompleted(sender As Object, e As EventArgs)
        Moving = False  ' 1枚目が開き終わった
    End Sub

    Private Sub MainWindow_Loaded(sender As Object, e As RoutedEventArgs) Handles Me.Loaded
        InitGame()
    End Sub

    Private Sub Startgame(sender As Object, e As RoutedEventArgs)

        For i As Integer = 0 To 15
            ' カードを裏返しにする
            OpenAnime(Grid1.FindName("Image" & i + 1), OpenMode.Close)
        Next
        Moving = False
        InitGame()

    End Sub
End Class
